@@ -39,24 +39,22 @@ LOGGER = logging.getLogger(__name__)
 # metabolic networks: Identification and removal. PLoS Computational
 # Biology, 13(4), 1–14. http://doi.org/10.1371/journal.pcbi.1005494
 ENERGY_COUPLES = {
-    'atp_c': 'adp_c',
-    'ctp_c': 'cdp_c',
-    'gtp_c': 'gdp_c',
-    'utp_c': 'udp_c',
-    'itp_c': 'idp_c',
-    'nadph_c': 'nadp_c',
-    'nadh_c': 'nad_c',
-    'fadh2_c': 'fad_c',
-    'fmnh2_c': 'fmn_c',
-    'q8h2_c': 'q8_c',
-    'mql8_c': 'mqn8_c',
-    'mql6_c': 'mqn6_c',
-    'mql7_c': 'mqn7_c',
-    '2dmmql8_c': '2dmmq8_c',
-    'accoa_c': 'coa_c',
-    'glu__L_c': 'akg_c',
-    'h_p': 'h_c'
-}
+    'MNXM3': 'MNXM7',
+    'MNXM63': 'MNXM220',
+    'MNXM51': 'MNXM30',
+    'MNXM121': 'MNXM17',
+    'MNXM423': 'MNXM495',
+    'MNXM6': 'MNXM5',
+    'MNXM10': 'MNXM8',
+    'MNXM38': 'MNXM33',
+    'MNXM208': 'MNXM119',
+    'MNXM191': 'MNXM232',
+    'MNXM223': 'MNXM509',
+    'MNXM7517': 'MNXM12235',
+    'MNXM12233': 'MNXM12236',
+    'MNXM558': 'MNXM2178',
+    'MNXM21': 'MNXM12',
+    'MNXM89557': 'MNXM20'}
 
 
 def check_stoichiometric_consistency(model):
@@ -267,6 +265,12 @@ def find_elementary_leakage_modes(model, atol=1e-13):
         "Coming soon™ if considered useful.")
 
 
+# TODO: I had to remove the test for an infinite proton gradient i.e testing
+# if the model can somehow cycle h_c and h_p with closed bounds. This test
+# however is very specfic for the existence of a periplasm and since it is
+# not trivial to presuppose agnostically between which two compartments a
+# given model builds the PMF, I've removed this capability from this test.
+# It should be re-added though once we can identify prokaryotes vs eukaryotes.
 def detect_energy_generating_cycles(model, metabolite_id):
     u"""
     Detect erroneous energy-generating cycles for a a single metabolite.
@@ -308,80 +312,86 @@ def detect_energy_generating_cycles(model, metabolite_id):
      Biology, 13(4), 1–14. http://doi.org/10.1371/journal.pcbi.1005494
 
     """
-    met = model.metabolites.get_by_id(metabolite_id)
-    dissipation_product = model.metabolites.get_by_id(ENERGY_COUPLES[met.id])
+    main_comp = helpers.find_compartment_id_in_model(model, 'c')
+    met = helpers.find_met_in_model(model, metabolite_id, main_comp)[0]
+
     dissipation_rxn = Reaction('Dissipation')
-    model = helpers.close_boundaries_sensibly(model)
-    model.add_reactions([dissipation_rxn])
-    if met.id in ['atp_c', 'ctp_c', 'gtp_c', 'utp_c', 'itp_c']:
+    if metabolite_id in ['MNXM3', 'MNXM63', 'MNXM51', 'MNXM121', 'MNXM423']:
         # build nucleotide-type dissipation reaction
-        dissipation_rxn.reaction = "h2o_c --> h_c + pi_c"
-    elif met.id in ['nadph_c', 'nadh_c']:
+        dissipation_rxn.add_metabolites({
+            helpers.find_met_in_model(model, 'MNXM2', main_comp)[0]: -1,
+            helpers.find_met_in_model(model, 'MNXM1', main_comp)[0]: 1,
+            helpers.find_met_in_model(model, 'MNXM9', main_comp)[0]: 1,
+        })
+    elif metabolite_id in ['MNXM6', 'MNXM10']:
         # build nicotinamide-type dissipation reaction
-        dissipation_rxn.reaction = "--> h_c"
-    elif met.id in ['fadh2_c', 'fmnh2_c', 'q8h2_c', 'mql8_c',
-                    'mql6_c', 'mql7_c', 'dmmql8_c']:
+        dissipation_rxn.add_metabolites({
+            helpers.find_met_in_model(model, 'MNXM1', main_comp)[0]: 1
+        })
+    elif metabolite_id in ['MNXM38', 'MNXM208', 'MNXM191', 'MNXM223',
+                           'MNXM7517', 'MNXM12233', 'MNXM558']:
         # build redox-partner-type dissipation reaction
-        dissipation_rxn.reaction = "--> 2 h_c"
-    elif met.id == 'accoa_c':
-        dissipation_rxn.reaction = "h2o_c --> h_c + ac_c"
-    elif met.id == 'glu__L_c':
-        dissipation_rxn.reaction = "h2o_c --> 2 h_c + nh3_c"
-    elif met.id == 'h_p':
-        pass
+        dissipation_rxn.add_metabolites({
+            helpers.find_met_in_model(model, 'MNXM1', main_comp)[0]: 2
+        })
+    elif metabolite_id == 'MNXM21':
+        dissipation_rxn.add_metabolites({
+            helpers.find_met_in_model(model, 'MNXM2', main_comp)[0]: -1,
+            helpers.find_met_in_model(model, 'MNXM1', main_comp)[0]: 1,
+            helpers.find_met_in_model(model, 'MNXM26', main_comp)[0]: 1,
+        })
+    elif metabolite_id == 'MNXM89557':
+        dissipation_rxn.add_metabolites({
+            helpers.find_met_in_model(model, 'MNXM2', main_comp)[0]: -1,
+            helpers.find_met_in_model(model, 'MNXM1', main_comp)[0]: 2,
+            helpers.find_met_in_model(model, 'MNXM15', main_comp)[0]: 1,
+        })
+
+    dissipation_product = helpers.find_met_in_model(
+        model,
+        ENERGY_COUPLES[metabolite_id],
+        main_comp)[0]
 
     dissipation_rxn.add_metabolites(
-        {met.id: -1, dissipation_product: 1})
+        {met: -1, dissipation_product: 1})
+    helpers.close_boundaries_sensibly(model)
+    model.add_reactions([dissipation_rxn])
     model.objective = dissipation_rxn
-    solution = model.optimize()
-    if solution.status == 'infeasible':
-        raise RuntimeError(
-            "The model cannot be solved as the solver status is"
-            "infeasible. This may be a bug."
-        )
-    elif solution.objective_value > 0.0:
+    solution = model.optimize(raise_error=True)
+
+    if solution.objective_value > 0.0:
         return solution.fluxes[solution.fluxes.abs() > 0.0].index. \
             drop(["Dissipation"]).tolist()
     else:
         return []
 
 
-def find_mass_imbalanced_reactions(model):
+def find_mass_unbalanced_reactions(reactions):
     """
     Find metabolic reactions that are not mass balanced.
 
-    This will exclude biomass, exchange and demand reactions as they are
-    unbalanced by definition. It will also fail all reactions where at
-    least one metabolite does not have a formula defined.
-
     Parameters
     ----------
-    model : cobra.Model
-        The metabolic model under investigation.
+    reactions : iterable
+        An iterable of ``cobra.Reaction``s.
 
     """
-    internal_rxns = con_helpers.get_internals(model)
     return [
-        rxn for rxn in internal_rxns if not con_helpers.is_mass_balanced(rxn)]
+        rxn for rxn in reactions if not con_helpers.is_mass_balanced(rxn)]
 
 
-def find_charge_imbalanced_reactions(model):
+def find_charge_unbalanced_reactions(reactions):
     """
     Find metabolic reactions that are not charge balanced.
 
-    This will exclude biomass, exchange and demand reactions as they are
-    unbalanced by definition. It will also fail all reactions where at
-    least one metabolite does not have a charge defined.
-
     Parameters
     ----------
-    model : cobra.Model
-        The metabolic model under investigation.
+    reactions : iterable
+        An iterable of ``cobra.Reaction``s.
 
     """
-    internal_rxns = con_helpers.get_internals(model)
     return [
-        rxn for rxn in internal_rxns if not con_helpers.is_charge_balanced(rxn)]
+        rxn for rxn in reactions if not con_helpers.is_charge_balanced(rxn)]
 
 
 def find_universally_blocked_reactions(model):
@@ -411,13 +421,9 @@ def find_universally_blocked_reactions(model):
 
 def find_stoichiometrically_balanced_cycles(model):
     u"""
-    Find metabolic rxns in stoichiometrically balanced cycles (SBCs).
+    Find metabolic reactions in stoichiometrically balanced cycles (SBCs).
 
-    The flux distribution of nominal FVA is compared with loopless FVA
-    (loopless=True) to determine reactions that participate in loops, as
-    participation in loops would increase the flux through a given reactions
-    to the maximal bounds. This function then returns reactions where the
-    flux differs between the two FVA calculations.
+    Identify forward and reverse cycles by closing all exchanges and using FVA.
 
     Parameters
     ----------
@@ -439,19 +445,10 @@ def find_stoichiometrically_balanced_cycles(model):
            http://doi.org/10.1038/nprot.2009.203
 
     """
-    try:
-        fva_result = flux_variability_analysis(model, loopless=False)
-        fva_result_loopless = flux_variability_analysis(model, loopless=True)
-    except Infeasible as err:
-        LOGGER.error("Failed to find stoichiometrically balanced cycles "
-                     "because '{}'. This may be a bug.".format(err))
-        return []
-    row_ids_max = fva_result[
-        fva_result.maximum != fva_result_loopless.maximum].index
-    row_ids_min = fva_result[
-        fva_result.minimum != fva_result_loopless.minimum].index
-    differential_fluxes = set(row_ids_min).union(set(row_ids_max))
-    return [model.reactions.get_by_id(id) for id in differential_fluxes]
+    helpers.close_boundaries_sensibly(model)
+    fva_result = flux_variability_analysis(model, loopless=False)
+    return fva_result.index[
+        (fva_result["minimum"] == -1) | (fva_result["maximum"] == 1)].tolist()
 
 
 def find_orphans(model):
@@ -510,14 +507,14 @@ def find_metabolites_produced_with_closed_bounds(model):
 
     """
     mets_produced = list()
-    model = helpers.close_boundaries_sensibly(model)
+    helpers.close_boundaries_sensibly(model)
     for met in model.metabolites:
         with model:
             exch = model.add_boundary(
                 met, type='irrex', reaction_id='IRREX', lb=0, ub=1
             )
             if helpers.run_fba(model, exch.id) > 0:
-                mets_produced.append(met.id)
+                mets_produced.append(met)
     return mets_produced
 
 
@@ -536,14 +533,14 @@ def find_metabolites_consumed_with_closed_bounds(model):
 
     """
     mets_consumed = list()
-    model = helpers.close_boundaries_sensibly(model)
+    helpers.close_boundaries_sensibly(model)
     for met in model.metabolites:
         with model:
             exch = model.add_boundary(
                 met, type='irrex', reaction_id='IRREX', lb=-1, ub=0
             )
             if helpers.run_fba(model, exch.id, direction='min') < 0:
-                mets_consumed.append(met.id)
+                mets_consumed.append(met)
     return mets_consumed
 
 
